'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var PropTypes = require('prop-types');
var cnb = _interopDefault(require('cnbuilder'));
var reactDraggable = require('react-draggable');

class RAFLoop {
    constructor() {
        /**
         * @description List of targets to update
         */
        this.targets = [];
        /**
         * @description ID of requested animation frame. Valuable only if loop is active and has items to iterate.
         */
        this.animationFrameID = 0;
        /**
         * @description Loop's state.
         */
        this._isActive = false;
        /**
         * @description Start the loop if it wasn't yet.
         */
        this.start = () => {
            if (!this._isActive && this.targets.length) {
                this._isActive = true;
                this.animationFrameID && cancelAnimationFrame(this.animationFrameID);
                this.animationFrameID = requestAnimationFrame(this.rafCallback);
            }
            return this;
        };
        /**
         * @description Stop the loop if is was active.
         */
        this.stop = () => {
            if (this._isActive) {
                this._isActive = false;
                this.animationFrameID && cancelAnimationFrame(this.animationFrameID);
                this.animationFrameID = 0;
            }
            return this;
        };
        /**
         * @description Callback that called each animation frame.
         */
        this.rafCallback = () => {
            if (!this._isActive) {
                return 0;
            }
            for (let i = 0; i < this.targets.length; i++) {
                !this.targets[i]._unmounted && this.targets[i].update();
            }
            return (this.animationFrameID = requestAnimationFrame(this.rafCallback));
        };
        /**
         * @description Add target to the iteration list if it's not there.
         */
        this.addTarget = (target, silent = false) => {
            if (this.targets.indexOf(target) === -1) {
                this.targets.push(target);
                this.targets.length === 1 && !silent && this.start();
            }
            return this;
        };
        /**
         * @description Remove target from iteration list if it was there.
         */
        this.removeTarget = (target) => {
            const idx = this.targets.indexOf(target);
            if (idx !== -1) {
                this.targets.splice(idx, 1);
                this.targets.length === 0 && this.stop();
            }
            return this;
        };
    }
    /**
     * @description Loop's state.
     */
    get isActive() {
        return this._isActive;
    }
}
var Loop = new RAFLoop();

var DIRECTION_AXIS;
(function (DIRECTION_AXIS) {
    DIRECTION_AXIS["X"] = "x";
    DIRECTION_AXIS["Y"] = "y";
})(DIRECTION_AXIS || (DIRECTION_AXIS = {}));
class ScrollbarTrack extends React.Component {
    constructor() {
        super(...arguments);
        this.element = null;
        this.ref = (ref) => {
            typeof this.props.elementRef === "function" && this.props.elementRef(ref);
            this.element = ref;
        };
        this.handleClick = (ev) => {
            if (!ev || !this.element || ev.button !== 0) {
                return;
            }
            if (typeof this.props.onClick === "function" &&
                ev.target === this.element) {
                if (typeof ev.offsetX !== "undefined") {
                    this.props.onClick(ev, {
                        axis: this.props.axis,
                        offset: this.props.axis === DIRECTION_AXIS.X ? ev.offsetX : ev.offsetY
                    });
                }
                else {
                    // support for old browsers
                    const rect = this.element.getBoundingClientRect();
                    this.props.onClick(ev, {
                        axis: this.props.axis,
                        offset: this.props.axis === DIRECTION_AXIS.X
                            ? ev.clientX - rect.left
                            : ev.clientY - rect.top
                    });
                }
            }
            return true;
        };
    }
    componentDidMount() {
        if (!this.element) {
            this.setState(() => {
                throw new Error("Element was not created. Possibly you haven't provided HTMLElement to renderer's `elementRef` function.");
            });
            return;
        }
        this.element.addEventListener("click", this.handleClick);
    }
    componentWillUnmount() {
        if (this.element) {
            this.element.removeEventListener("click", this.handleClick);
            this.element = null;
        }
    }
    render() {
        const { elementRef, renderer, axis, onClick, ...props } = this.props;
        props.className = cnb("ScrollbarTrack", axis === DIRECTION_AXIS.X ? "ScrollbarTrack-X" : "ScrollbarTrack-Y", props.className);
        return renderer ? (renderer({
            ...props,
            axis,
            elementRef: this.ref
        })) : (React.createElement("div", Object.assign({}, props, { ref: this.ref })));
    }
}
ScrollbarTrack.propTypes = {
    axis: PropTypes.oneOf([DIRECTION_AXIS.X, DIRECTION_AXIS.Y]),
    className: PropTypes.string,
    style: PropTypes.object,
    onClick: PropTypes.func,
    elementRef: PropTypes.func,
    renderer: PropTypes.func
};

class ScrollbarThumb extends React.Component {
    constructor() {
        super(...arguments);
        this.element = null;
        this.initialOffsetX = 0;
        this.initialOffsetY = 0;
        this.lastDragData = {
            x: 0,
            y: 0,
            deltaX: 0,
            deltaY: 0,
            lastX: 0,
            lastY: 0
        };
        this.handleOnDragStart = (ev, data) => {
            if (!this.element) {
                this.handleOnDragStop(ev, data);
                return;
            }
            if (global.document) {
                this.prevUserSelect = global.document.body.style.userSelect;
                global.document.body.style.userSelect = "none";
                //@ts-ignore
                this.prevOnSelectStart = global.document.onselectstart;
                //@ts-ignore
                global.document.onselectstart = ScrollbarThumb.selectStartReplacer;
            }
            this.props.onDragStart &&
                this.props.onDragStart((this.lastDragData = {
                    x: data.x - this.initialOffsetX,
                    y: data.y - this.initialOffsetY,
                    lastX: data.lastX - this.initialOffsetX,
                    lastY: data.lastY - this.initialOffsetY,
                    deltaX: data.deltaX,
                    deltaY: data.deltaY
                }));
            this.element.classList.add("dragging");
        };
        this.handleOnDrag = (ev, data) => {
            if (!this.element) {
                this.handleOnDragStop(ev, data);
                return;
            }
            this.props.onDrag &&
                this.props.onDrag((this.lastDragData = {
                    x: data.x - this.initialOffsetX,
                    y: data.y - this.initialOffsetY,
                    lastX: data.lastX - this.initialOffsetX,
                    lastY: data.lastY - this.initialOffsetY,
                    deltaX: data.deltaX,
                    deltaY: data.deltaY
                }));
        };
        this.handleOnDragStop = (ev, data) => {
            const resultData = data
                ? {
                    x: data.x - this.initialOffsetX,
                    y: data.y - this.initialOffsetY,
                    lastX: data.lastX - this.initialOffsetX,
                    lastY: data.lastY - this.initialOffsetY,
                    deltaX: data.deltaX,
                    deltaY: data.deltaY
                }
                : this.lastDragData;
            this.props.onDragEnd && this.props.onDragEnd(resultData);
            this.element && this.element.classList.remove("dragging");
            if (global.document) {
                global.document.body.style.userSelect = this.prevUserSelect;
                this.prevUserSelect = null;
                // @ts-ignore
                global.document.onselectstart = this.prevOnSelectStart;
                this.prevOnSelectStart = null;
            }
            this.initialOffsetX = 0;
            this.initialOffsetY = 0;
            this.lastDragData = {
                x: 0,
                y: 0,
                deltaX: 0,
                deltaY: 0,
                lastX: 0,
                lastY: 0
            };
        };
        this.handleOnMouseDown = (ev) => {
            if (!this.element) {
                return;
            }
            ev.stopPropagation();
            if (typeof ev.offsetX !== "undefined") {
                this.initialOffsetX = ev.offsetX;
                this.initialOffsetY = ev.offsetY;
            }
            else {
                const rect = this.element.getBoundingClientRect();
                this.initialOffsetX = ev.clientX - rect.left;
                this.initialOffsetY = ev.clientY - rect.top;
            }
        };
        this.ref = (ref) => {
            typeof this.props.elementRef === "function" && this.props.elementRef(ref);
            this.element = ref;
        };
    }
    componentDidMount() {
        if (!this.element) {
            this.setState(() => {
                throw new Error("Element was not created. Possibly you haven't provided HTMLElement to renderer's `elementRef` function.");
            });
            return;
        }
    }
    componentWillUnmount() {
        this.handleOnDragStop();
    }
    render() {
        const { elementRef, renderer, axis, onClick, onDrag, onDragEnd, onDragStart, ...props } = this.props;
        props.className = cnb("ScrollbarThumb", axis === DIRECTION_AXIS.X ? "ScrollbarThumb-X" : "ScrollbarThumb-Y", props.className);
        return (React.createElement(reactDraggable.DraggableCore, { allowAnyClick: false, enableUserSelectHack: false, onMouseDown: this.handleOnMouseDown, onDrag: this.handleOnDrag, onStart: this.handleOnDragStart, onStop: this.handleOnDragStop }, renderer ? (renderer({
            ...props,
            axis,
            elementRef: this.ref
        })) : (React.createElement("div", Object.assign({}, props, { ref: this.ref })))));
    }
}
ScrollbarThumb.propTypes = {
    axis: PropTypes.oneOf([DIRECTION_AXIS.X, DIRECTION_AXIS.Y]),
    className: PropTypes.string,
    style: PropTypes.object,
    onClick: PropTypes.func,
    elementRef: PropTypes.func,
    renderer: PropTypes.func
};
ScrollbarThumb.selectStartReplacer = () => false;

let doc = global.document || null;
let isReverseRTLScrollNeeded = null;
/**
 * @description Detect need of horizontal scroll reverse while RTL
 */
const shouldReverseRTLScroll = (force = false) => {
    if (!force && isReverseRTLScrollNeeded !== null) {
        return isReverseRTLScrollNeeded;
    }
    if (!doc) {
        return (isReverseRTLScrollNeeded = false);
    }
    let el = doc.createElement("div");
    let child = doc.createElement("div");
    el.setAttribute("style", "display:block;position:absolute;width:100px;height:100px;top:-9999px;overflow:scroll;direction:rtl;");
    child.setAttribute("style", "display:block;position:relative;width:1000px;height:1000px;direction:rtl;");
    el.appendChild(child);
    doc.body.appendChild(el);
    el.scrollLeft;
    el.scrollLeft = 45;
    isReverseRTLScrollNeeded = el.scrollLeft === 0;
    doc.body.removeChild(el);
    return isReverseRTLScrollNeeded;
};
/**
 * @description Return element's height without padding
 *
 * ts-ignore here is okay here, because it brigs around 40% of performance
 */
const getInnerHeight = (el) => {
    const styles = getComputedStyle(el);
    if (styles.boxSizing === "border-box") {
        return Math.max(
        // @ts-ignore
        (parseFloat(styles.height) || 0) -
            // @ts-ignore
            (parseFloat(styles.paddingTop) || 0) -
            // @ts-ignore
            (parseFloat(styles.paddingBottom) || 0), 0);
    }
    // @ts-ignore
    return parseFloat(styles.height) || 0;
};
/**
 * @description Return element's width without padding
 *
 * ts-ignore here is okay here, because it brigs around 40% of performance
 */
const getInnerWidth = (el) => {
    const styles = getComputedStyle(el);
    if (styles.boxSizing === "border-box") {
        return Math.max(
        // @ts-ignore
        (parseFloat(styles.width) || 0) -
            // @ts-ignore
            (parseFloat(styles.paddingLeft) || 0) -
            // @ts-ignore
            (parseFloat(styles.paddingRight) || 0), 0);
    }
    // @ts-ignore
    return parseFloat(styles.width) || 0;
};
/**
 * @description Return unique UUID v4
 */
const uuid = () => {
    let uuid = "";
    for (let i = 0; i < 32; i++) {
        if (i === 8 || i === 20) {
            uuid += "-" + ((Math.random() * 16) | 0).toString(16);
        }
        else if (i === 12) {
            uuid += "-4";
        }
        else if (i === 16) {
            uuid += "-" + ((Math.random() * 16) | (0 & 3) | 8).toString(16);
        }
        else {
            uuid += ((Math.random() * 16) | 0).toString(16);
        }
    }
    return uuid;
};
/**
 * @description Calculate thumb size for given viewport and track parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} minimalSize - Minimal thumb's size
 * @param {number} maximalSize - Maximal thumb's size
 */
function calcThumbSize(contentSize, viewportSize, trackSize, minimalSize, maximalSize) {
    if (viewportSize >= contentSize) {
        return 0;
    }
    let thumbSize = (viewportSize / contentSize) * trackSize;
    if (typeof maximalSize === "number") {
        thumbSize = Math.min(maximalSize, thumbSize);
    }
    if (typeof minimalSize === "number") {
        thumbSize = Math.max(minimalSize, thumbSize);
    }
    return thumbSize;
}
/**
 * @description Calculate thumb offset for given viewport, track and thumb parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} thumbSize - Thumb size
 * @param {number} scroll - Scroll value to represent
 */
function calcThumbOffset(contentSize, viewportSize, trackSize, thumbSize, scroll) {
    if (!scroll || !thumbSize || viewportSize >= contentSize) {
        return 0;
    }
    return ((trackSize - thumbSize) * scroll) / (contentSize - viewportSize);
}
/**
 * @description Calculate scroll for given viewport, track and thumb parameters
 *
 * @param {number} contentSize - Scrollable content size
 * @param {number} viewportSize - Viewport size
 * @param {number} trackSize - Track size thumb can move
 * @param {number} thumbSize - Thumb size
 * @param {number} thumbOffset - Thumb's offset representing the scroll
 */
function calcScrollForThumbOffset(contentSize, viewportSize, trackSize, thumbSize, thumbOffset) {
    if (!thumbOffset || !thumbSize || viewportSize >= contentSize) {
        return 0;
    }
    return (thumbOffset * (contentSize - viewportSize)) / (trackSize - thumbSize);
}
let scrollbarWidth = null;
/**
 * @description Returns scrollbar width specific for current environment
 */
function getScrollbarWidth(force = false) {
    if (!force && scrollbarWidth !== null) {
        return scrollbarWidth;
    }
    if (!doc) {
        return (scrollbarWidth = 0);
    }
    let el = doc.createElement("div");
    el.setAttribute("style", "display:block;position:absolute;width:100px;height:100px;top:-9999px;overflow:scroll;");
    doc.body.appendChild(el);
    scrollbarWidth = el.offsetWidth - el.clientWidth;
    doc.body.removeChild(el);
    return scrollbarWidth;
}

class Emittr {
    constructor(maxHandlers = 10) {
        this.setMaxHandlers(maxHandlers);
        this._handlers = Object.create(null);
    }
    setMaxHandlers(count) {
        if (typeof count !== "number" || count < 0 || !count) {
            throw new TypeError(`Expected maxHandlers to be a positive number, got '${count}' of type ${typeof count}`);
        }
        this._maxHandlers = count;
        return this;
    }
    getMaxHandlers() {
        return this._maxHandlers;
    }
    emit(name, ...args) {
        if (typeof this._handlers[name] !== "object" ||
            !Array.isArray(this._handlers[name])) {
            return false;
        }
        Emittr._callEventHandlers(this, this._handlers[name], args);
        return true;
    }
    static _callEventHandlers(emitter, handlers, args) {
        if (!handlers.length) {
            return;
        }
        if (handlers.length === 1) {
            Reflect.apply(handlers[0], emitter, args);
            return;
        }
        handlers = [...handlers];
        let idx;
        for (idx = 0; idx < handlers.length; idx++) {
            Reflect.apply(handlers[idx], emitter, args);
        }
    }
    _wrapOnceHandler(name, handler) {
        const onceState = {
            fired: false,
            handler,
            wrappedHandler: undefined,
            emitter: this,
            event: name
        };
        const wrappedHandler = Emittr._onceWrapper.bind(onceState);
        onceState.wrappedHandler = wrappedHandler;
        wrappedHandler.handler = handler;
        wrappedHandler.event = name;
        return wrappedHandler;
    }
    on(name, handler) {
        Emittr._addHandler(this, name, handler);
        return this;
    }
    prependOn(name, handler) {
        Emittr._addHandler(this, name, handler, true);
        return this;
    }
    once(name, handler) {
        if (typeof handler !== "function") {
            throw new TypeError("Expected event handler to be a function, got " + typeof handler);
        }
        Emittr._addHandler(this, name, this._wrapOnceHandler(name, handler));
        return this;
    }
    prependOnce(name, handler) {
        if (typeof handler !== "function") {
            throw new TypeError("Expected event handler to be a function, got " + typeof handler);
        }
        Emittr._addHandler(this, name, this._wrapOnceHandler(name, handler), true);
        return this;
    }
    off(name, handler) {
        Emittr._removeHandler(this, name, handler);
        return this;
    }
    removeAllHandlers() {
        const handlers = this._handlers;
        this._handlers = Object.create(null);
        const removeHandlers = handlers["removeHandler"];
        delete handlers["removeHandler"];
        let idx, eventName;
        for (eventName in handlers) {
            for (idx = handlers[eventName].length - 1; idx >= 0; idx--) {
                Emittr._callEventHandlers(this, removeHandlers, [
                    eventName,
                    handlers[eventName][idx].handler ||
                        handlers[eventName][idx]
                ]);
            }
        }
        return this;
    }
}
Emittr._addHandler = (emitter, name, handler, prepend = false) => {
    if (typeof handler !== "function") {
        throw new TypeError("Expected event handler to be a function, got " + typeof handler);
    }
    emitter._handlers[name] = emitter._handlers[name] || [];
    emitter.emit("addHandler", name, handler);
    prepend
        ? emitter._handlers[name].unshift(handler)
        : emitter._handlers[name].push(handler);
    return emitter;
};
Emittr._onceWrapper = function _onceWrapper(...args) {
    if (!this.fired) {
        this.fired = true;
        this.emitter.off(this.event, this.wrappedHandler);
        Reflect.apply(this.handler, this.emitter, args);
    }
};
Emittr._removeHandler = (emitter, name, handler) => {
    if (typeof handler !== "function") {
        throw new TypeError("Expected event handler to be a function, got " + typeof handler);
    }
    if (typeof emitter._handlers[name] === "undefined" ||
        !emitter._handlers[name].length) {
        return emitter;
    }
    let idx = -1;
    if (emitter._handlers[name].length === 1) {
        if (emitter._handlers[name][0] === handler ||
            emitter._handlers[name][0].handler === handler) {
            idx = 0;
            handler =
                emitter._handlers[name][0].handler ||
                    emitter._handlers[name][0];
        }
    }
    else {
        for (idx = emitter._handlers[name].length - 1; idx >= 0; idx--) {
            if (emitter._handlers[name][idx] === handler ||
                emitter._handlers[name][idx].handler === handler) {
                handler =
                    emitter._handlers[name][idx].handler ||
                        emitter._handlers[name][idx];
                break;
            }
        }
    }
    if (idx === -1) {
        return emitter;
    }
    idx === 0
        ? emitter._handlers[name].shift()
        : emitter._handlers[name].splice(idx, 1);
    emitter.emit("removeHandler", name, handler);
    return emitter;
};

const reverseRTL = shouldReverseRTLScroll();
var SCROLLBAR_TRACK_CLICK_BEHAVIOR;
(function (SCROLLBAR_TRACK_CLICK_BEHAVIOR) {
    SCROLLBAR_TRACK_CLICK_BEHAVIOR["JUMP"] = "jump";
    SCROLLBAR_TRACK_CLICK_BEHAVIOR["STEP"] = "step";
})(SCROLLBAR_TRACK_CLICK_BEHAVIOR || (SCROLLBAR_TRACK_CLICK_BEHAVIOR = {}));
const ScrollbarContext = React.createContext({ parentScrollbar: null });
class Scrollbar extends React.Component {
    constructor(props) {
        super(props);
        /**
         * @description Get current scroll-related values.<br/>
         * If <i>force</i> if truthy - will recalculate them instead of returning cached values.
         *
         * @return ScrollValues
         */
        this.getScrollValues = (force = false) => {
            if (this.scrollValues && !force) {
                return { ...this.scrollValues };
            }
            let scrollValues = {
                clientHeight: 0,
                clientWidth: 0,
                scrollHeight: 0,
                scrollWidth: 0,
                scrollTop: 0,
                scrollLeft: 0,
                scrollYBlocked: false,
                scrollXBlocked: false,
                scrollYPossible: false,
                scrollXPossible: false,
                trackYVisible: false,
                trackXVisible: false,
                isRTL: undefined
            };
            scrollValues.isRTL = this.state.isRTL;
            scrollValues.scrollYBlocked = this.props.noScroll || this.props.noScrollY;
            scrollValues.scrollXBlocked = this.props.noScroll || this.props.noScrollX;
            if (this.contentElement) {
                scrollValues.clientHeight = this.contentElement.clientHeight;
                scrollValues.clientWidth = this.contentElement.clientWidth;
                scrollValues.scrollHeight = this.contentElement.scrollHeight;
                scrollValues.scrollWidth = this.contentElement.scrollWidth;
                scrollValues.scrollTop = this.contentElement.scrollTop;
                scrollValues.scrollLeft = this.contentElement.scrollLeft;
                scrollValues.scrollYPossible =
                    !scrollValues.scrollYBlocked &&
                        scrollValues.scrollHeight > scrollValues.clientHeight;
                scrollValues.scrollXPossible =
                    !scrollValues.scrollXBlocked &&
                        scrollValues.scrollWidth > scrollValues.clientWidth;
                scrollValues.trackYVisible =
                    scrollValues.scrollYPossible ||
                        this.props.permanentTracks ||
                        this.props.permanentTrackY;
                scrollValues.trackXVisible =
                    scrollValues.scrollXPossible ||
                        this.props.permanentTracks ||
                        this.props.permanentTrackX;
            }
            return scrollValues;
        };
        /**
         * @description Scroll to top border
         */
        this.scrollToTop = () => {
            if (this.contentElement) {
                this.contentElement.scrollTop = 0;
            }
            return this;
        };
        /**
         * @description Scroll to left border
         */
        this.scrollToLeft = () => {
            if (this.contentElement) {
                this.contentElement.scrollLeft = 0;
            }
            return this;
        };
        /**
         * @description Scroll to bottom border
         */
        this.scrollToBottom = () => {
            if (this.contentElement) {
                this.contentElement.scrollTop =
                    this.contentElement.scrollHeight - this.contentElement.clientHeight;
            }
            return this;
        };
        /**
         * @description Scroll to right border
         */
        this.scrollToRight = () => {
            if (this.contentElement) {
                this.contentElement.scrollLeft =
                    this.contentElement.scrollWidth - this.contentElement.clientWidth;
            }
            return this;
        };
        /**
         * @description Set the scrolls at given coordinates.<br/>
         * If coordinate is undefined - current scroll value will persist.
         */
        this.scrollTo = (x, y) => {
            if (this.contentElement) {
                typeof x === "number" && (this.contentElement.scrollLeft = x);
                typeof y === "number" && (this.contentElement.scrollTop = y);
            }
            return this;
        };
        /**
         * @description Center the viewport at given coordinates.<br/>
         * If coordinate is undefined - current scroll value will persist.
         */
        this.centerAt = (x, y) => {
            if (this.contentElement) {
                typeof x === "number" &&
                    (this.contentElement.scrollLeft =
                        x - this.contentElement.clientWidth / 2);
                typeof y === "number" &&
                    (this.contentElement.scrollTop =
                        y - this.contentElement.clientHeight / 2);
            }
            return this;
        };
        this.update = (force = false) => {
            if (!this.contentElement) {
                return;
            }
            // autodetect direction if not defined
            if (typeof this.state.isRTL === "undefined") {
                this.setState({
                    isRTL: getComputedStyle(this.contentElement).direction === "rtl"
                });
                return this.getScrollValues();
            }
            const scrollValues = this.getScrollValues(true);
            const prevScrollValues = { ...this.scrollValues };
            let bitmask = 0;
            if (!force) {
                prevScrollValues.clientHeight !== scrollValues.clientHeight &&
                    (bitmask |= 1 << 0);
                prevScrollValues.clientWidth !== scrollValues.clientWidth &&
                    (bitmask |= 1 << 1);
                prevScrollValues.scrollHeight !== scrollValues.scrollHeight &&
                    (bitmask |= 1 << 2);
                prevScrollValues.scrollWidth !== scrollValues.scrollWidth &&
                    (bitmask |= 1 << 3);
                prevScrollValues.scrollTop !== scrollValues.scrollTop &&
                    (bitmask |= 1 << 4);
                prevScrollValues.scrollLeft !== scrollValues.scrollLeft &&
                    (bitmask |= 1 << 5);
                prevScrollValues.scrollYBlocked !== scrollValues.scrollYBlocked &&
                    (bitmask |= 1 << 6);
                prevScrollValues.scrollXBlocked !== scrollValues.scrollXBlocked &&
                    (bitmask |= 1 << 7);
                prevScrollValues.scrollYPossible !== scrollValues.scrollYPossible &&
                    (bitmask |= 1 << 8);
                prevScrollValues.scrollXPossible !== scrollValues.scrollXPossible &&
                    (bitmask |= 1 << 9);
                prevScrollValues.trackYVisible !== scrollValues.trackYVisible &&
                    (bitmask |= 1 << 10);
                prevScrollValues.trackXVisible !== scrollValues.trackXVisible &&
                    (bitmask |= 1 << 11);
                prevScrollValues.isRTL !== scrollValues.isRTL && (bitmask |= 1 << 12);
                // if not forced and nothing has changed - skip this update
                if (bitmask === 0) {
                    return prevScrollValues;
                }
            }
            else {
                bitmask = 0b1111111111111;
            }
            // if scrollbars visibility has changed
            if (bitmask & (1 << 10) || bitmask & (1 << 11)) {
                prevScrollValues.scrollYBlocked = scrollValues.scrollYBlocked;
                prevScrollValues.scrollXBlocked = scrollValues.scrollXBlocked;
                prevScrollValues.scrollYPossible = scrollValues.scrollYPossible;
                prevScrollValues.scrollXPossible = scrollValues.scrollXPossible;
                if (this.trackYElement && bitmask & (1 << 10)) {
                    this.trackYElement.style.display = scrollValues.trackYVisible
                        ? null
                        : "none";
                }
                if (this.trackXElement && bitmask & (1 << 11)) {
                    this.trackXElement.style.display = scrollValues.trackXVisible
                        ? null
                        : "none";
                }
                this.scrollValues = prevScrollValues;
                this.setState({
                    trackYVisible: (this.scrollValues.trackYVisible =
                        scrollValues.trackYVisible),
                    trackXVisible: (this.scrollValues.trackXVisible =
                        scrollValues.trackXVisible)
                });
                return;
            }
            (this.props.native ? this.updaterNative : this.updaterCustom)(bitmask, scrollValues);
            this.scrollValues = scrollValues;
            this.eventEmitter.emit("update", { ...scrollValues }, prevScrollValues);
            (bitmask & (1 << 4) || bitmask & (1 << 5)) &&
                this.eventEmitter.emit("scroll", { ...scrollValues }, prevScrollValues);
            return this.scrollValues;
        };
        this.updaterNative = () => {
            // just for future
            return true;
        };
        this.updaterCustom = (bitmask, scrollValues) => {
            if (this.trackYElement) {
                if (this.thumbYElement &&
                    (bitmask & (1 << 0) ||
                        bitmask & (1 << 2) ||
                        bitmask & (1 << 4) ||
                        bitmask & (1 << 6) ||
                        bitmask & (1 << 8))) {
                    if (scrollValues.scrollYPossible) {
                        const trackInnerSize = getInnerHeight(this.trackYElement);
                        const thumbSize = calcThumbSize(scrollValues.scrollHeight, scrollValues.clientHeight, trackInnerSize, this.props.minimalThumbYSize || this.props.minimalThumbSize, this.props.maximalThumbYSize || this.props.maximalThumbSize);
                        const thumbOffset = calcThumbOffset(scrollValues.scrollHeight, scrollValues.clientHeight, trackInnerSize, thumbSize, scrollValues.scrollTop);
                        this.thumbYElement.style.transform = `translateY(${thumbOffset}px)`;
                        this.thumbYElement.style.height = thumbSize + "px";
                        this.thumbYElement.style.display = null;
                    }
                    else {
                        this.thumbYElement.style.transform = null;
                        this.thumbYElement.style.height = "0px";
                        this.thumbYElement.style.display = "none";
                    }
                }
            }
            if (this.trackXElement) {
                if (this.thumbXElement &&
                    (bitmask & (1 << 1) ||
                        bitmask & (1 << 3) ||
                        bitmask & (1 << 5) ||
                        bitmask & (1 << 7) ||
                        bitmask & (1 << 9) ||
                        bitmask & (1 << 12))) {
                    if (scrollValues.scrollXPossible) {
                        const trackInnerSize = getInnerWidth(this.trackXElement);
                        const thumbSize = calcThumbSize(scrollValues.scrollWidth, scrollValues.clientWidth, trackInnerSize, this.props.minimalThumbXSize || this.props.minimalThumbSize, this.props.maximalThumbXSize || this.props.maximalThumbSize);
                        let thumbOffset = calcThumbOffset(scrollValues.scrollWidth, scrollValues.clientWidth, trackInnerSize, thumbSize, scrollValues.scrollLeft);
                        if (reverseRTL) {
                            thumbOffset += trackInnerSize - thumbSize;
                        }
                        this.thumbXElement.style.transform = `translateX(${thumbOffset}px)`;
                        this.thumbXElement.style.width = thumbSize + "px";
                        this.thumbXElement.style.display = null;
                    }
                    else {
                        this.thumbXElement.style.transform = null;
                        this.thumbXElement.style.width = "0px";
                        this.thumbXElement.style.display = "none";
                    }
                }
            }
            return true;
        };
        this.elementRefHolder = (ref) => {
            this.holderElement = ref;
            typeof this.props.elementRef === "function" && this.props.elementRef(ref);
        };
        this.elementRefWrapper = (ref) => {
            this.wrapperElement = ref;
            typeof this.props.wrapperProps.elementRef === "function" &&
                this.props.wrapperProps.elementRef(ref);
        };
        this.elementRefContent = (ref) => {
            this.contentElement = ref;
            typeof this.props.contentProps.elementRef === "function" &&
                this.props.contentProps.elementRef(ref);
        };
        this.elementRefTrackX = (ref) => {
            this.trackXElement = ref;
            typeof this.props.trackXProps.elementRef === "function" &&
                this.props.trackXProps.elementRef(ref);
        };
        this.elementRefTrackY = (ref) => {
            this.trackYElement = ref;
            typeof this.props.trackYProps.elementRef === "function" &&
                this.props.trackYProps.elementRef(ref);
        };
        this.elementRefThumbX = (ref) => {
            this.thumbXElement = ref;
            typeof this.props.thumbXProps.elementRef === "function" &&
                this.props.thumbXProps.elementRef(ref);
        };
        this.elementRefThumbY = (ref) => {
            this.thumbYElement = ref;
            typeof this.props.thumbYProps.elementRef === "function" &&
                this.props.thumbYProps.elementRef(ref);
        };
        this.handleTrackXClick = (ev, values) => {
            this.props.trackXProps.onClick &&
                this.props.trackXProps.onClick(ev, values);
            if (!this.contentElement ||
                !this.trackXElement ||
                !this.thumbXElement ||
                !this.scrollValues ||
                !this.scrollValues.scrollXPossible) {
                return;
            }
            this._scrollDetection();
            const thumbSize = this.thumbXElement.clientWidth;
            const trackInnerSize = getInnerWidth(this.trackXElement);
            const thumbOffset = (this.scrollValues.isRTL && reverseRTL
                ? values.offset + thumbSize / 2 - trackInnerSize
                : values.offset - thumbSize / 2) -
                //@ts-ignore
                (parseFloat(getComputedStyle(this.trackXElement).paddingLeft) || 0);
            let target = calcScrollForThumbOffset(this.scrollValues.scrollWidth, this.scrollValues.clientWidth, trackInnerSize, thumbSize, thumbOffset);
            if (this.props.trackClickBehavior === SCROLLBAR_TRACK_CLICK_BEHAVIOR.STEP) {
                target = (this.scrollValues.isRTL
                    ? this.scrollValues.scrollLeft > target
                    : this.scrollValues.scrollLeft < target)
                    ? this.scrollValues.scrollLeft + this.scrollValues.clientWidth
                    : this.scrollValues.scrollLeft - this.scrollValues.clientWidth;
            }
            this.contentElement.scrollLeft = target;
        };
        this.handleTrackYClick = (ev, values) => {
            this.props.trackYProps.onClick &&
                this.props.trackYProps.onClick(ev, values);
            if (!this.contentElement ||
                !this.trackYElement ||
                !this.thumbYElement ||
                !this.scrollValues ||
                !this.scrollValues.scrollYPossible) {
                return;
            }
            this._scrollDetection();
            const thumbSize = this.thumbYElement.clientHeight;
            let target = calcScrollForThumbOffset(this.scrollValues.scrollHeight, this.scrollValues.clientHeight, getInnerHeight(this.trackYElement), thumbSize, values.offset - thumbSize / 2) -
                //@ts-ignore
                (parseFloat(getComputedStyle(this.trackYElement).paddingTop) || 0);
            if (this.props.trackClickBehavior === SCROLLBAR_TRACK_CLICK_BEHAVIOR.JUMP) {
                this.contentElement.scrollTop = target;
            }
            else {
                this.contentElement.scrollTop =
                    this.scrollValues.scrollTop < target
                        ? this.scrollValues.scrollTop + this.scrollValues.clientHeight
                        : this.scrollValues.scrollTop - this.scrollValues.clientHeight;
            }
        };
        this.handleTrackYMouseWheel = (ev) => {
            this.props.trackYProps &&
                this.props.trackYProps.onWheel &&
                this.props.trackYProps.onWheel(ev);
            this._scrollDetection();
            if (!this.contentElement || this.scrollValues.scrollYBlocked) {
                return;
            }
            this.scrollTop += ev.deltaY;
        };
        this.handleTrackXMouseWheel = (ev) => {
            this.props.trackXProps &&
                this.props.trackXProps.onWheel &&
                this.props.trackXProps.onWheel(ev);
            this._scrollDetection();
            if (!this.contentElement || this.scrollValues.scrollXBlocked) {
                return;
            }
            this.scrollLeft += ev.deltaX;
        };
        this.handleThumbXDrag = (data) => {
            if (!this.trackXElement ||
                !this.thumbXElement ||
                !this.contentElement ||
                !this.scrollValues ||
                !this.scrollValues.scrollXPossible) {
                return;
            }
            this._scrollDetection();
            const trackRect = this.trackXElement.getBoundingClientRect();
            const styles = getComputedStyle(this.trackXElement);
            //@ts-ignore
            const paddingLeft = parseFloat(styles.paddingLeft) || 0;
            //@ts-ignore
            const paddingRight = parseFloat(styles.paddingRight) || 0;
            const trackInnerSize = trackRect.width - paddingLeft - paddingRight;
            const thumbSize = this.thumbXElement.clientWidth;
            const offset = this.scrollValues.isRTL && reverseRTL
                ? data.x + thumbSize - trackInnerSize + paddingLeft
                : data.lastX - paddingLeft;
            this.contentElement.scrollLeft = calcScrollForThumbOffset(this.scrollValues.scrollWidth, this.scrollValues.clientWidth, trackInnerSize, thumbSize, offset);
        };
        this.handleThumbYDrag = (data) => {
            if (!this.contentElement ||
                !this.trackYElement ||
                !this.thumbYElement ||
                !this.scrollValues ||
                !this.scrollValues.scrollYPossible) {
                return;
            }
            this._scrollDetection();
            const trackRect = this.trackYElement.getBoundingClientRect();
            const styles = getComputedStyle(this.trackYElement);
            //@ts-ignore
            const paddingTop = parseFloat(styles.paddingTop) || 0;
            //@ts-ignore
            const paddingBottom = parseFloat(styles.paddingBottom) || 0;
            const trackInnerSize = trackRect.height - paddingTop - paddingBottom;
            const thumbSize = this.thumbYElement.clientHeight;
            const offset = data.y - paddingTop;
            this.contentElement.scrollTop = calcScrollForThumbOffset(this.scrollValues.scrollHeight, this.scrollValues.clientHeight, trackInnerSize, thumbSize, offset);
        };
        this.handleContentScroll = () => {
            this._scrollDetection();
        };
        this._scrollDetection = () => {
            !this._scrollDetectionTO &&
                this.eventEmitter.emit("scrollStart", this.getScrollValues());
            this._scrollDetectionTO && global.clearTimeout(this._scrollDetectionTO);
            this._scrollDetectionTO = global.setTimeout(this._scrollDetectionCallback, this.props.scrollDetectionThreshold || 0);
        };
        this._scrollDetectionCallback = () => {
            this._scrollDetectionTO = null;
            this.eventEmitter.emit("scrollStop", this.getScrollValues());
        };
        this.state = {
            trackXVisible: false,
            trackYVisible: false,
            isRTL: this.props.rtl
        };
        this.scrollValues = this.getScrollValues(true);
        this.eventEmitter = new Emittr(15);
        this.props.onUpdate && this.eventEmitter.on("update", this.props.onUpdate);
        this.props.onScroll && this.eventEmitter.on("scroll", this.props.onScroll);
        this.props.onScrollStart &&
            this.eventEmitter.on("scrollStart", this.props.onScrollStart);
        this.props.onScrollEnd &&
            this.eventEmitter.on("scrollEnd", this.props.onScrollEnd);
        this.id = uuid();
    }
    componentDidMount() {
        if (!this.contentElement) {
            this.setState(() => {
                throw new Error("content element was not created. Possibly you haven't provided HTMLElement to renderer's `elementRef` function.");
            });
            return;
        }
        if (!this.props.native) {
            if (!this.holderElement) {
                this.setState(() => {
                    throw new Error("holder element was not created. Possibly you haven't provided HTMLElement to renderer's `elementRef` function.");
                });
                return;
            }
            if (!this.wrapperElement) {
                this.setState(() => {
                    throw new Error("wrapper element was not created. Possibly you haven't provided HTMLElement to renderer's `elementRef` function.");
                });
                return;
            }
        }
        Loop.addTarget(this);
        if (typeof this.props.scrollTop !== "undefined") {
            this.contentElement.scrollTop = this.props.scrollTop;
        }
        if (typeof this.props.scrollLeft !== "undefined") {
            this.contentElement.scrollLeft = this.props.scrollLeft;
        }
        this.update(true);
    }
    componentWillUnmount() {
        Loop.removeTarget(this);
    }
    componentDidUpdate(prevProps, prevState) {
        if (!this.contentElement) {
            return;
        }
        if (this.props.rtl !== prevProps.rtl &&
            this.props.rtl !== this.state.isRTL) {
            this.setState({ isRTL: this.props.rtl });
        }
        if (this.state.isRTL !== prevState.isRTL) {
            this.update();
        }
        if (typeof this.props.scrollTop !== "undefined" &&
            this.props.scrollTop !== this.contentElement.scrollTop) {
            this.contentElement.scrollTop = this.props.scrollTop;
        }
        if (typeof this.props.scrollLeft !== "undefined" &&
            this.props.scrollLeft !== this.contentElement.scrollLeft) {
            this.contentElement.scrollLeft = this.props.scrollLeft;
        }
        if (prevProps.onUpdate !== this.props.onUpdate) {
            prevProps.onUpdate && this.eventEmitter.off("update", prevProps.onUpdate);
            this.props.onUpdate &&
                this.eventEmitter.on("update", this.props.onUpdate);
        }
        if (prevProps.onScroll !== this.props.onScroll) {
            prevProps.onScroll && this.eventEmitter.off("scroll", prevProps.onScroll);
            this.props.onScroll &&
                this.eventEmitter.on("scroll", this.props.onScroll);
        }
        if (prevProps.onScrollStart !== this.props.onScrollStart) {
            prevProps.onScrollStart &&
                this.eventEmitter.off("scrollStart", prevProps.onScrollStart);
            this.props.onScrollStart &&
                this.eventEmitter.on("scrollStart", this.props.onScrollStart);
        }
        if (prevProps.onScrollEnd !== this.props.onScrollEnd) {
            prevProps.onScrollEnd &&
                this.eventEmitter.off("scrollEnd", prevProps.onScrollEnd);
            this.props.onScrollEnd &&
                this.eventEmitter.on("scrollEnd", this.props.onScrollEnd);
        }
    }
    get scrollTop() {
        if (this.contentElement) {
            return this.contentElement.scrollTop;
        }
        return 0;
    }
    set scrollTop(top) {
        if (this.contentElement) {
            this.contentElement.scrollTop = top;
            this.update();
        }
    }
    get scrollLeft() {
        if (this.contentElement) {
            return this.contentElement.scrollLeft;
        }
        return 0;
    }
    set scrollLeft(left) {
        if (this.contentElement) {
            this.contentElement.scrollLeft = left;
        }
    }
    get scrollHeight() {
        if (this.contentElement) {
            return this.contentElement.scrollHeight;
        }
        return 0;
    }
    get scrollWidth() {
        if (this.contentElement) {
            return this.contentElement.scrollWidth;
        }
        return 0;
    }
    get clientHeight() {
        if (this.contentElement) {
            return this.contentElement.clientHeight;
        }
        return 0;
    }
    get clientWidth() {
        if (this.contentElement) {
            return this.contentElement.clientWidth;
        }
        return 0;
    }
    render() {
        const { createContext, rtl, native, momentum, noDefaultStyles, noScrollX, noScrollY, noScroll, permanentTrackX, permanentTrackY, permanentTracks, removeTracksWhenNotUsed, removeTrackYWhenNotUsed, removeTrackXWhenNotUsed, minimalThumbSize, maximalThumbSize, minimalThumbXSize, maximalThumbXSize, minimalThumbYSize, maximalThumbYSize, fallbackScrollbarWidth, trackClickBehavior, scrollDetectionThreshold, wrapperProps: propsWrapperProps, contentProps: propsContentProps, trackXProps: propsTrackXProps, trackYProps: propsTrackYProps, thumbXProps: propsThumbXProps, thumbYProps: propsThumbYProps, scrollbarWidth: propsScrollbarWidth, elementRef, renderer, onScroll, children, ...props } = this.props;
        if (native) {
            this.elementRefHolder(null);
            this.elementRefWrapper(null);
            this.elementRefTrackX(null);
            this.elementRefTrackY(null);
            this.elementRefThumbX(null);
            this.elementRefThumbY(null);
            return (React.createElement("div", { key: "ScrollbarCustom", className: cnb("ScrollbarsCustom native", {
                    trackYVisible: this.state.trackYVisible,
                    trackXVisible: this.state.trackXVisible,
                    rtl: this.state.isRTL
                }, this.props.className), style: {
                    ...this.props.style,
                    ...(rtl && { direction: "rtl" }),
                    ...(momentum && { WebkitOverflowScrolling: "touch" }),
                    overflowX: noScroll || noScrollX
                        ? "hidden"
                        : permanentTracks || permanentTrackX
                            ? "scroll"
                            : "auto",
                    overflowY: noScroll || noScrollY
                        ? "hidden"
                        : permanentTracks || permanentTrackY
                            ? "scroll"
                            : "auto"
                }, onScroll: this.handleContentScroll, children: children, ref: this.elementRefContent }));
        }
        const styles = Scrollbar.calculateStyles(this.props, this.state, this.scrollValues, typeof propsScrollbarWidth !== "undefined"
            ? propsScrollbarWidth
            : getScrollbarWidth());
        const contentProps = {
            ...propsContentProps,
            key: "ScrollbarCustom-content",
            className: cnb("ScrollbarContent", propsContentProps.className),
            style: styles.content,
            onScroll: this.handleContentScroll,
            children: createContext ? (React.createElement(ScrollbarContext.Provider, { value: { parentScrollbar: this } }, children)) : (children)
        };
        if (propsContentProps.renderer) {
            delete contentProps.ref;
            contentProps.elementRef = this.elementRefContent;
        }
        else {
            delete contentProps.elementRef;
            contentProps.ref = this.elementRefContent;
        }
        const wrapperProps = {
            ...propsWrapperProps,
            key: "ScrollbarCustom-wrapper",
            className: cnb("ScrollbarWrapper", propsWrapperProps.className),
            style: styles.wrapper,
            children: propsContentProps.renderer ? (propsContentProps.renderer(contentProps)) : (React.createElement("div", Object.assign({}, contentProps)))
        };
        if (propsWrapperProps.renderer) {
            delete wrapperProps.ref;
            wrapperProps.elementRef = this.elementRefWrapper;
        }
        else {
            delete wrapperProps.elementRef;
            wrapperProps.ref = this.elementRefWrapper;
        }
        const holderProps = {
            ...props,
            key: "ScrollbarCustom-holder_" + this.id,
            className: cnb("ScrollbarsCustom", {
                trackYVisible: this.state.trackYVisible,
                trackXVisible: this.state.trackXVisible,
                rtl: this.state.isRTL
            }, this.props.className),
            style: styles.holder
        };
        if (renderer) {
            delete holderProps.ref;
            holderProps.elementRef = this.elementRefHolder;
        }
        else {
            delete holderProps.elementRef;
            holderProps.ref = this.elementRefHolder;
        }
        const holderChildren = [
            propsWrapperProps.renderer ? (propsWrapperProps.renderer(wrapperProps)) : (React.createElement("div", Object.assign({}, wrapperProps)))
        ];
        if (this.state.trackYVisible ||
            (!removeTracksWhenNotUsed && !removeTrackYWhenNotUsed)) {
            const thumbYProps = {
                ...propsThumbYProps,
                key: "ScrollbarCustom-thumbY",
                style: styles.thumbY,
                elementRef: this.elementRefThumbY,
                onDrag: this.handleThumbYDrag,
                onDragEnd: this.handleThumbYDrag
            };
            const trackYProps = {
                ...propsTrackYProps,
                key: "ScrollbarCustom-trackY",
                style: styles.trackY,
                elementRef: this.elementRefTrackY,
                onClick: this.handleTrackYClick,
                onWheel: this.handleTrackYMouseWheel,
                children: (React.createElement(ScrollbarThumb, Object.assign({ axis: DIRECTION_AXIS.Y }, thumbYProps, { ref: undefined })))
            };
            holderChildren.push(React.createElement(ScrollbarTrack, Object.assign({ axis: DIRECTION_AXIS.Y }, trackYProps, { ref: undefined })));
        }
        else {
            this.elementRefTrackY(null);
            this.elementRefThumbY(null);
        }
        if (this.state.trackXVisible ||
            (!removeTracksWhenNotUsed && !removeTrackXWhenNotUsed)) {
            const thumbXProps = {
                ...propsThumbXProps,
                key: "ScrollbarCustom-thumbX",
                style: styles.thumbX,
                elementRef: this.elementRefThumbX,
                onDrag: this.handleThumbXDrag,
                onDragEnd: this.handleThumbXDrag
            };
            const trackXProps = {
                ...propsTrackXProps,
                key: "ScrollbarCustom-trackX",
                style: styles.trackX,
                elementRef: this.elementRefTrackX,
                onClick: this.handleTrackXClick,
                onWheel: this.handleTrackXMouseWheel,
                children: (React.createElement(ScrollbarThumb, Object.assign({ axis: DIRECTION_AXIS.X }, thumbXProps, { ref: undefined })))
            };
            holderChildren.push(React.createElement(ScrollbarTrack, Object.assign({ axis: DIRECTION_AXIS.X }, trackXProps, { ref: undefined })));
        }
        else {
            this.elementRefTrackX(null);
            this.elementRefThumbX(null);
        }
        holderProps.children = holderChildren;
        return renderer ? renderer(holderProps) : React.createElement("div", Object.assign({}, holderProps));
    }
    static calculateStyles(props, state, scrollValues, scrollbarWidth) {
        const useDefaultStyles = !props.noDefaultStyles;
        return {
            holder: {
                ...(useDefaultStyles && defaultStyles.holder),
                ...props.style
            },
            wrapper: {
                ...(useDefaultStyles && {
                    ...defaultStyles.wrapper,
                    [state.isRTL ? "left" : "right"]: state.trackYVisible ? 10 : 0,
                    bottom: state.trackXVisible ? 10 : 0
                }),
                ...props.wrapperProps.style,
                position: "absolute",
                overflow: "hidden"
            },
            content: {
                ...(useDefaultStyles && defaultStyles.content),
                ...props.contentProps.style,
                ...(typeof props.rtl !== "undefined" && {
                    direction: props.rtl ? "rtl" : "ltr"
                }),
                ...(props.momentum && { WebkitOverflowScrolling: "touch" }),
                overflowY: scrollValues.scrollYPossible ? "scroll" : "hidden",
                overflowX: scrollValues.scrollXPossible ? "scroll" : "hidden",
                paddingBottom: !scrollbarWidth && scrollValues.scrollXPossible
                    ? props.fallbackScrollbarWidth
                    : undefined,
                marginBottom: scrollValues.scrollXPossible
                    ? -(scrollbarWidth || props.fallbackScrollbarWidth)
                    : undefined,
                [state.isRTL ? "paddingLeft" : "paddingRight"]: !scrollbarWidth && scrollValues.scrollYPossible
                    ? props.fallbackScrollbarWidth
                    : undefined,
                [state.isRTL
                    ? "marginLeft"
                    : "marginRight"]: scrollValues.scrollYPossible
                    ? -(scrollbarWidth || props.fallbackScrollbarWidth)
                    : undefined
            },
            trackX: {
                ...(useDefaultStyles && defaultStyles.track.common),
                ...(useDefaultStyles && defaultStyles.track.x),
                ...props.trackXProps.style,
                ...(!state.trackXVisible && { display: "none" })
            },
            trackY: {
                ...(useDefaultStyles && defaultStyles.track.common),
                ...(useDefaultStyles && defaultStyles.track.y),
                ...props.trackYProps.style,
                [state.isRTL ? "left" : "right"]: 0,
                ...(!state.trackYVisible && { display: "none" })
            },
            thumbX: {
                ...(useDefaultStyles && defaultStyles.thumb.common),
                ...(useDefaultStyles && defaultStyles.thumb.x),
                ...props.thumbXProps.style
            },
            thumbY: {
                ...(useDefaultStyles && defaultStyles.thumb.common),
                ...(useDefaultStyles && defaultStyles.thumb.y),
                ...props.thumbYProps.style
            }
        };
    }
}
Scrollbar.contextType = ScrollbarContext;
Scrollbar.propTypes = {
    createContext: PropTypes.bool,
    rtl: PropTypes.bool,
    native: PropTypes.bool,
    momentum: PropTypes.bool,
    noDefaultStyles: PropTypes.bool,
    minimalThumbSize: PropTypes.number,
    maximalThumbSize: PropTypes.number,
    minimalThumbXSize: PropTypes.number,
    maximalThumbXSize: PropTypes.number,
    minimalThumbYSize: PropTypes.number,
    maximalThumbYSize: PropTypes.number,
    noScrollX: PropTypes.bool,
    noScrollY: PropTypes.bool,
    noScroll: PropTypes.bool,
    permanentTrackX: PropTypes.bool,
    permanentTrackY: PropTypes.bool,
    permanentTracks: PropTypes.bool,
    removeTracksWhenNotUsed: PropTypes.bool,
    removeTrackYWhenNotUsed: PropTypes.bool,
    removeTrackXWhenNotUsed: PropTypes.bool,
    trackClickBehavior: PropTypes.oneOf([
        SCROLLBAR_TRACK_CLICK_BEHAVIOR.JUMP,
        SCROLLBAR_TRACK_CLICK_BEHAVIOR.STEP
    ]),
    scrollbarWidth: PropTypes.number,
    fallbackScrollbarWidth: PropTypes.number,
    scrollDetectionThreshold: PropTypes.number,
    scrollTop: PropTypes.number,
    scrollLeft: PropTypes.number,
    className: PropTypes.string,
    wrapperProps: PropTypes.object,
    contentProps: PropTypes.object,
    trackXProps: PropTypes.object,
    trackYProps: PropTypes.object,
    thumbXProps: PropTypes.object,
    thumbYProps: PropTypes.object,
    onUpdate: PropTypes.func,
    onScroll: PropTypes.func,
    onScrollStart: PropTypes.func,
    onScrollEnd: PropTypes.func
};
Scrollbar.defaultProps = {
    createContext: false,
    native: false,
    momentum: true,
    noDefaultStyles: false,
    noScrollX: false,
    noScrollY: false,
    noScroll: false,
    permanentTrackX: false,
    permanentTrackY: false,
    permanentTracks: false,
    removeTracksWhenNotUsed: false,
    removeTrackYWhenNotUsed: false,
    removeTrackXWhenNotUsed: false,
    minimalThumbSize: 30,
    fallbackScrollbarWidth: 20,
    trackClickBehavior: SCROLLBAR_TRACK_CLICK_BEHAVIOR.JUMP,
    scrollDetectionThreshold: 100,
    wrapperProps: {},
    contentProps: {},
    trackXProps: {},
    trackYProps: {},
    thumbXProps: {},
    thumbYProps: {}
};
const defaultStyles = {
    holder: {
        position: "relative",
        width: "100%",
        height: "100%"
    },
    wrapper: {
        position: "absolute",
        top: 0,
        left: 0,
        bottom: 0,
        right: 0
    },
    content: {
        position: "absolute",
        top: 0,
        left: 0,
        bottom: 0,
        right: 0
    },
    track: {
        common: {
            position: "absolute",
            overflow: "hidden",
            borderRadius: 4,
            background: "rgba(0,0,0,.1)",
            userSelect: "none"
        },
        x: {
            height: 10,
            width: "calc(100% - 20px)",
            bottom: 0,
            left: 10
        },
        y: {
            width: 10,
            height: "calc(100% - 20px)",
            top: 10
        }
    },
    thumb: {
        common: {
            cursor: "pointer",
            borderRadius: 4,
            background: "rgba(0,0,0,.4)"
        },
        x: {
            height: "100%",
            width: 0
        },
        y: {
            width: "100%",
            height: 0
        }
    }
};

exports.default = Scrollbar;
exports.Scrollbar = Scrollbar;
exports.ScrollbarContext = ScrollbarContext;
